import java.util.*;


public class Metro { 
    int       maxCrosses;
    Hashtable stations;

    public int getNumberOfStations() {
        return stations.size();
    }
    
    public Metro(Line[] lines, Cross[] crosses) { 
        stations = new Hashtable();
        for (int i = 0; i < lines.length; i++) {
            Line line = lines[i];
            Station[] lineStations = line.stations;
            for (int j = 0; j < lineStations.length; j++) { 
                lineStations[j].line = line;
                stations.put(lineStations[j].name, lineStations[j]);
            }
            for (int j = 1; j < lineStations.length; j++) { 
                lineStations[j-1].addEdge(lineStations[j], lineStations[j-1].distance);
                lineStations[j].addEdge(lineStations[j-1], lineStations[j-1].distance);
            }
            if (lineStations[lineStations.length-1].distance != 0) { 
                lineStations[0].addEdge(lineStations[lineStations.length-1], 
                                        lineStations[lineStations.length-1].distance);
                lineStations[lineStations.length-1].addEdge(lineStations[0], 
                                                            lineStations[lineStations.length-1].distance);
            }
        }
        maxCrosses = crosses.length; 
        for (int i = 0; i < crosses.length; i++) { 
            Cross cross = crosses[i];
            Station s1 = (Station)stations.get(cross.station1);
            Station s2 = (Station)stations.get(cross.station2);
            if (s1 == null) { 
                throw new RuntimeException("No such station for cross " + i + " " + cross.station1);
            }
            if (s2 == null) { 
                throw new RuntimeException("No such station for cross " + i + " " + cross.station2);
            }
            s1.addCrossEdge(s2, cross.time);
            s2.addCrossEdge(s1, cross.time);
        }
    }
    
    Station findStation(String name) { 
        return (Station)stations.get(name);
    }

    Route findRoute(Station from, Station till, int crossPenalty) { 
        // Graph is simple - do not use dynamic programming
        // and use the simplest recursion traversal
        Route route = new Route();
        route.crosses = new CrossEdge[0];
        if (from == till) { 
            return route;
        }
        Enumeration e = stations.elements();
        while (e.hasMoreElements()) {
            ((Station)e.nextElement()).reset();
        }   
        from.traverse(till, 0, crossPenalty);
        CrossEdge[] crosses = new CrossEdge[maxCrosses];
        int nCrosses = 0;
        int nStations = 0;
        int time = 0;
        Edge edge = from.optimal; 
        while (true) { 
            if (edge.isCross()) { 
                crosses[nCrosses++] = (CrossEdge)edge;
            } else { 
                nStations += 1;
            }
            time += edge.time;
            if (edge.dst == till) { 
                break;
            }
            edge = edge.dst.optimal;
        }
        route.stations = new Station[nStations+nCrosses*2+1];
        route.stations[0] = from;
        edge = from.optimal; 
        int i = 0;
        while (true) { 
            if (edge.isCross()) { 
                route.stations[++i] = ((CrossEdge)edge).src;
            } 
            route.stations[++i] = edge.dst;
            if (edge.dst == till) { 
                break;
            }
            edge = edge.dst.optimal;
        }
        route.time = time;
        route.nStations = nStations;
        if (nCrosses > 0) { 
            route.crosses = new CrossEdge[nCrosses];
            System.arraycopy(crosses,0, route.crosses, 0, nCrosses);
        }
        return route;
    }

    void printRoute(Route route) { 
        System.out.println("Time: " + route.time);
        System.out.println("Station: " + route.nStations);
        for (int i = 0; i < route.crosses.length; i++) { 
            String from = route.crosses[i].src.getShortName();
            String to = route.crosses[i].dst.getShortName();
            if (from.equals(to)) { 
                System.out.println("Cross at station " + from
                                   + " of line " + route.crosses[i].src.line.name
                                   + " to line " + route.crosses[i].dst.line.name);
            } else {
                System.out.println("Cross from station " + from
                                   + " of line " + route.crosses[i].src.line.name
                                   + " to line " + route.crosses[i].dst.line.name);
            }
        }
    }        
}

